from matplotlib import pyplot as plt
import matplotlib.patches as patches

from torchvision import transforms
from torch import optim
import torch.nn.functional as F

import math
import numpy as np

from tqdm.notebook import tqdm
import pandas as pd
from skimage import io
import torch

from pathlib import Path
import requests


class DotaDataset_solution():
    def __init__(self,
                 dataset_df,
                 transforms,
                 label_map,
                 images_ram,
                 output_img_size=(64, 64, 3),
                 ):

        
        # 
        self.samples_df = dataset_df     # dataframe of all samples
        self.label_map = label_map       # maping of string labels to integer values
        
        # transforms
        self.transform = transforms      # list of transforms, similar to built-in datasets

        
        # pointer to ram memmory
        self.images_ram = images_ram      
        
        
    def __len__(self):
        return len(self.samples_df)
    
    def __getitem__(self, index):
        
        # get sample atributes
        sample = self.samples_df.iloc[index]
        img_path = sample['img_path']
        
        # get label
        label = self.label_map[sample.label]
        
        # load image from the RAM
        img = self.images_ram[img_path]
        
        # locate object
        meanr, meanc = np.mean([sample.minr, sample.maxr], dtype=int), np.mean([sample.minc, sample.maxc], dtype=int)
        size = int(np.maximum(sample.maxr - sample.minr + 1, sample.maxc - sample.minc + 1) * 1.4 )
        
        min_row_bbox = max(meanr - size//2, 0)
        min_col_bbox = max(meanc - size//2, 0)
        max_row_bbox = meanr + size//2
        max_col_bbox = meanc + size//2


        # crop image
        img = img[min_row_bbox:max_row_bbox, min_col_bbox:max_col_bbox]
                
        # apply transformations
        img = self.transform(image=img)['image']
        
        return img, label

########### TUTORIAL TOOLS ################

def download_dota(fpath, fname, url):
    zip_file_path = fpath / fname

    if not (zip_file_path).exists():
        
        print(f'downloading ... {url + fname}')
        response = requests.get(url + fname, stream=True)
        total_length = response.headers.get('content-length')
        with open(zip_file_path, 'wb') as zip_file:
            if total_length is None: # no content length header
                zip_file.write(response.content)
            else:
                total_length = int(total_length)
                total_its = total_length // 4096
                for data in tqdm(response.iter_content(chunk_size=4096), total=total_its):
                    zip_file.write(data)

class RAMDiskForImages:
    '''
    Loads images to RAM to faster the access.
    '''
    
    def __init__(self):
        self.images = {}
        
    def __getitem__(self, img_path):
        '''
        Returns an image stored at 'image_path'
        '''
        
        return self.images[img_path]
    
    def load_from_dataframe(self, dataset_df):
        '''
        Reads DataFrame 'dataset_df' and loads all the images
        mentioned in the 'image_path' column.
        '''
        
        img_paths = dataset_df['img_path'].unique()
        
        print('Loading image dataset to RAM memory...')
        for img_path in tqdm(img_paths):
            if img_path not in self.images.keys():

                self.images[img_path] = io.imread(img_path)


def ishow(img,
          cmap='viridis',
          title='',
          fig_size=(8,6),
          colorbar=False,
          interpolation='none'):
    ' Function `ishow` displays an image in a new window. '
    
    #extent = (0, img.shape[1], img.shape[0], 0)
    fig, ax = plt.subplots(figsize=fig_size)
    pcm = ax.imshow(img,
              #extent=extent,
              cmap=cmap,
              interpolation=interpolation)
    
    ax.set_frame_on(False)
    plt.title(title)
    plt.tight_layout()
    if colorbar:
        fig.colorbar(pcm, orientation='vertical')
    plt.show()
    
    
def plot_losses(losses):
    
    plt.figure()
    for i, (loss, val_loss) in enumerate(losses):
        plt.plot(loss, c=f'C{i}', linestyle='--')
        plt.plot(val_loss, c=f'C{i}')
                
    plt.legend(['loss - AUG',
               'validation loss - AUG',
               'loss',
               'validation loss'])
    plt.show()
    
    
class DeviceDataLoader:
    def __init__(self, dl, dev):
        self.dl = dl
        self.dev = dev

    def __len__(self):
        return len(self.dl)

    def __iter__(self):
        batches = iter(self.dl)
        for x, y in batches:

            yield x.to(self.dev), y.to(self.dev)
    
    
def loss_batch(model, loss_func, xb, yb, opt=None):
    loss = loss_func(model(xb), yb)

    if opt is not None:
        loss.backward()
        opt.step()
        opt.zero_grad()

    return loss.item()
    
    
    
def train_model(epochs, model, loss_func, train_dl, valid_dl, lr=0.001):
    '''
    train function from tutorial 02
    
    '''
    import logging
    import traceback
    try:
        print('Starting training')
        opt = optim.Adam(model.parameters(), lr=lr)
        
        model.eval()
        curr_loss = 0
        n_samples = 0
        for xb, yb in tqdm(train_dl, total=len(train_dl)):
            curr_loss += loss_batch(model, loss_func, xb, yb, opt=opt) * len(xb)
            n_samples += len(xb)
        with torch.no_grad():

            val_losses, nums = zip(
                *[(loss_batch(model, loss_func, xb, yb), len(xb)) for xb, yb in valid_dl]
            )
            val_loss = np.sum(np.multiply(val_losses, nums)) / np.sum(nums)
        loss_log, loss_val_log = [curr_loss / n_samples], [np.sum(np.multiply(val_losses, nums)) / np.sum(nums)]
        model.zero_grad()
        for epoch in tqdm(range(epochs), 'training'):
            
            curr_loss = 0
            n_samples = 0
            
            model.train()
            for xb, yb in tqdm(train_dl, total=len(train_dl)):
                curr_loss += loss_batch(model, loss_func, xb, yb, opt=opt) * len(xb)
                n_samples += len(xb)
                
            loss = curr_loss / n_samples
            loss_log.append(loss)
    
            model.eval()
            with torch.no_grad():
                val_losses, nums = zip(
                    *[(loss_batch(model, loss_func, xb, yb), len(xb)) for xb, yb in valid_dl]
                )
            val_loss = np.sum(np.multiply(val_losses, nums)) / np.sum(nums)
            loss_val_log.append(val_loss)
    
            print(f'epoch {epoch + 1}/{epochs}, loss: {loss:.05f}, validation loss: {val_loss:.05f}')
        
        return loss_log, loss_val_log
    except Exception as e:
        logging.error(traceback.format_exc())


def read_DOTA_labels(labels_path):
    '''
    Reads labels of DOTA dataset in the original format.
    Returns labels in a Dataframe.
    '''
    
    # get all files
    label_files = [file for file in labels_path.iterdir()]
    
    # iterate over label files
    samples = []
    for file_path in label_files:
        
        # get img path
        img_path = file_path.parents[1] / 'images' / Path(file_path.stem + '.png') 
        
        if not img_path.exists():
            continue
            
        assert img_path.exists(), img_path
        #img = io.imread(img_path)
        #w, h = img.shape[0], img.shape[1]
        
        with open(file_path, 'r') as f:
            
            for line in f.readlines():
                d = line.rstrip().split(' ')
                if len(d) != 10:
                    continue
                    
                # read sample
                columns = [int(float(d[i])) for i in [0, 2, 4, 6]]
                rows = [int(float(d[i])) for i in [1, 3, 5, 7]]
                label = d[8]
                
                # make a bbox alligned to the axes
                minr, maxr = min(rows), max(rows)
                minc, maxc = min(columns), max(columns)
                
                assert minr < maxr
                assert minc < maxc
                
                ''' slows down the 
                # include only annotations inside the image domain 
                if (minr > w) or (minc > h):
                    print('INFO: annotation outside the image domain was removed - ', file_path, '|', line.rstrip())
                    continue
                '''
  
                samples.append((str(img_path), minr, maxr, minc, maxc, label))

    return pd.DataFrame(samples, columns=['img_path', 'minr', 'maxr', 'minc', 'maxc', 'label'])     


def show_sample_DOTA(df, sample_idx):
    
    sample = df.iloc[sample_idx]
    img = io.imread(sample.img_path)
    
    # Create a Rectangle patch
    fig, ax = plt.subplots(figsize=(11,8))

    # Display the image
    ax.imshow(img)

    # Create a Rectangle patch
    rect = patches.Rectangle((sample.minc, sample.minr),
                             sample.maxc-sample.minc,
                             sample.maxr-sample.minr,
                             linewidth=1,
                             edgecolor='r',
                             facecolor='none')

    # Add the patch to the Axes
    ax.add_patch(rect)
    ax.set_title(sample.label)
    plt.show()


def show_augmentation(ds, idx, n_samples=3):
    '''
    shows grid of NxN images
    '''
    
    # create figure
    figure = plt.figure(figsize=(8, 8))
    
    
    cols, rows = n_samples, n_samples
    for i in range(1, cols * rows + 1):
        img, label = ds[idx]
        
        #img, label = ds[idx]
        #img = img['image']
        
        
        if img.min() < 0 or img.max() > 1:
            
            print(f'The image values were shifted to the range (0, 1). The original range is ({img.min():.04f}), {img.max():.04f})')
        
            img -= img.min()
            img /= img.max()
        
        figure.add_subplot(rows, cols, i)
        plt.axis("off")
        plt.imshow(img.permute(1, 2, 0).cpu().detach().numpy())
    plt.suptitle(f'sample no {idx}, label {label}')
    plt.show()
    
    
def show_sample(ds, idx):
    '''
    shows ine image sample at index 'idx'
    '''
    
    # create figure
    figure = plt.figure(figsize=(5, 5))
    img, label = ds[idx]

    if img.min() < 0 or img.max() > 1:
        
        img = img.float()

        print(f'The image values were shifted to the range (0, 1). The original range is ({img.min():.04f}, {img.max():.04f})')

        img -= img.min()
        img /= img.max()

    plt.imshow(img.permute(1, 2, 0).cpu().detach().numpy())
    plt.suptitle(f'sample no.: {idx}, label: {label}')
    plt.show()
    
    
class Small_CNN(torch.nn.Module):
    def __init__(self, input_shape=(32, 32, 3), out_channels=10):
        super().__init__()
        
        self.W, self.H, self.D = input_shape
        self.conv1 = torch.nn.Conv2d(self.D, 32, kernel_size=3, stride=2, padding=1) # 16x16x32
        self.conv2 = torch.nn.Conv2d(32, 64, kernel_size=3, stride=2, padding=1) # 8x8x64
        self.conv3 = torch.nn.Conv2d(64, out_channels, kernel_size=3, stride=2, padding=1) # 4x4x15
        
        self.pool_coef = int(input_shape[0] // 8)

    def forward(self, xb):
        
        xb = xb.view(-1, self.D, self.W, self.H)
        
        xb = F.relu(self.conv1(xb))
        xb = F.relu(self.conv2(xb))
        xb = F.relu(self.conv3(xb))
        xb = F.avg_pool2d(xb, self.pool_coef)
        xb = xb.view(-1, xb.size(1))
        
        
        return xb

def test_model(model, data, classes):
    model.eval()
    curr_loss = 0
    n_samples = 0
    acc = 0
    confusion = np.zeros((len(classes), len(classes)))

    
    y_pred = []
    y_true = []
    
    # iterate over test data
    for xb, yb in tqdm(data, total=len(data)):
            output = model(xb) 
            output = output.exp()
            output = (torch.argmax(output/output.sum(), 1)).data.cpu().numpy()
            
            y_pred.extend(output) # Save Prediction
            
            yb = yb.data.cpu().numpy()
            y_true.extend(yb) # Save Truth

            acc += np.sum(output == yb)
    acc = acc/len(data)
    for yp, yt in zip(y_pred, y_true):
            confusion[yp, yt] += 1
    return confusion, acc

def print_conf(confusion, labels):
    confusion = confusion.astype(np.uint64)
    print(f'X\t', end='')
    for j in range(confusion.shape[1]):
        print(f'|{j}\t', end='')
    print('\n', end='')
    for i in range(confusion.shape[0]):
        print(f'{i}\t', end='')
        for j in range(confusion.shape[1]):
            print(f'|{confusion[i, j]}\t', end='')
        print('\n', end='')
        
